/*
 * Copyright (c) 2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Wed Jun 24 2015
 */
#include <errno.h>
#include <fnmatch.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include "libtk3.h"


/* --- local data ---------------------------------------------------------- */

static struct tk3_channel_s *channels;
static int nchannels;

pthread_mutex_t chanex = PTHREAD_MUTEX_INITIALIZER;
pthread_t chanreader;

static const char *devmagic[] = {
  [TK3_MKBL] = "[?]?mkbl[0-9].[0-9]",
  [TK3_MKFL] = "[?]mkfl[0-9].[0-9]"
};

static void *	tk3_reader(void *data);


/* --- tk3_client ---------------------------------------------------------- */

int
tk3_client(enum tk3_device dev, const char *path, uint32_t speed,
           char *emsg, size_t emsgsize)
{
  struct tk3_channel_s chan = { .fd = -1 }, *nc;
  struct stat sb;
  int i, s, c;

  chan.fd = tk3_open_tty(path, speed);
  if (chan.fd < 0) goto syserror;

  if (fstat(chan.fd, &sb)) goto syserror;

  /* look for existing channel */
  pthread_mutex_lock(&chanex);
  for(i = 0; i < nchannels; i++) {
    if (channels[i].fd >= 0 &&
        channels[i].st_dev == sb.st_dev &&
        channels[i].st_ino == sb.st_ino) {
      if (dev == channels[i].data.dev)
        channels[i].nrefs++;
      else i = -1;
      pthread_mutex_unlock(&chanex);
      close(chan.fd);
      if (i >= 0) return i;
      errno = EBADF;
      goto syserror;
    }
  }

  /* test connection */
  while (tk3_recv_msg(&chan, true) == 1); /* flush buffer */
  c = 0;
  do {
    do {
      if (tk3_send_msg(&chan, "?")) /* ask for id */ {
        pthread_mutex_unlock(&chanex);
        goto syserror;
      }

      s = tk3_wait_msg(&(struct tk3_channel_s *){&chan}, 1, &chanex);
      if (s < 0 && errno != EINTR) {
        pthread_mutex_unlock(&chanex);
        goto syserror;
      }
      if (s > 0) break;
    } while(c++ < 3);

    s = tk3_recv_msg(&chan, true);
  } while(s == 1 && chan.msg[0] != '?');

  if (s != 1 || fnmatch(devmagic[dev], (char *)chan.msg, 0)) {
    pthread_mutex_unlock(&chanex);
    errno = EBADF;
    goto syserror;
  }

  switch(dev) {
    case TK3_MKBL:
      chan.data.mkbl.id = chan.msg[1];
      break;
    case TK3_MKFL:
      chan.data.mkfl.motor = NULL;
      chan.data.mkfl.nmotor = 0;
      break;
    case TK3_NONE: break;
  }

  /* create channel */
  nc = realloc(channels, (1 + nchannels) * sizeof(*channels));
  if (!nc) {
    pthread_mutex_unlock(&chanex);
    goto syserror;
  }
  channels = nc;

  if (!nchannels) {
    if (pthread_create(&chanreader, NULL, tk3_reader, NULL)) {
      pthread_mutex_unlock(&chanex);
      goto syserror;
    }
  }

  chan.data.dev = dev;
  strncpy(chan.path, path, sizeof(chan.path));
  chan.st_dev = sb.st_dev;
  chan.st_ino = sb.st_ino;
  chan.nrefs = 1;
  chan.start = 0;
  chan.escape = 0;
  chan.len = 0;
  channels[nchannels++] = chan;

  pthread_mutex_unlock(&chanex);
  return nchannels-1;

syserror:
  snprintf(emsg, emsgsize, "%s: %s", path, strerror(errno));
  if (chan.fd >= 0) close(chan.fd);
  return -1;
}

static inline struct tk3_channel_s *
tk3_channel(int c, enum tk3_device dev, char *emsg, size_t emsgsize)
{
  if (c >= 0 && c < nchannels) {
    if (dev == TK3_NONE || channels[c].data.dev == dev)
      return channels + c;

    snprintf(emsg, emsgsize, "%s: bad device", channels[c].path);
    return NULL;
  }

  snprintf(emsg, emsgsize, "%d: bad channel", c);
  return NULL;
}


/* --- tk3_close ----------------------------------------------------------- */

int
tk3_close(int channel, char *emsg, size_t emsgsize)
{
  int i, r;
  struct tk3_channel_s *c = tk3_channel(channel, TK3_NONE, emsg, emsgsize);
  if (!c) return -1;

  if (--c->nrefs) return 0;

  (void)tk3_send_data(channel, "stop", NULL, 0, emsg, emsgsize);
  switch(c->data.dev) {
    case TK3_MKBL:
      (void)tk3_rqst_data(channel, "motor", 0, emsg, emsgsize);
      break;

    case TK3_MKFL:
      (void)tk3_rqst_data(channel, "imu", 0, emsg, emsgsize);
      (void)tk3_rqst_data(channel, "motor", 0, emsg, emsgsize);
      break;

    case TK3_NONE: break;
  }

  pthread_mutex_lock(&chanex);
  if (c->fd >=0) close(c->fd);
  c->fd = -1;

  for(i=r=0; i<nchannels; i++)
    if (channels[i].fd >=0) r++;
  if (r) {
    pthread_mutex_unlock(&chanex);
    return 0;
  }

  nchannels = 0;
  if (channels) free(channels);
  channels = NULL;
  pthread_mutex_unlock(&chanex);

  pthread_join(chanreader, NULL);
  return 0;
}


/* --- tk3_data ------------------------------------------------------------ */

struct tk3_data
tk3_data(int channel, char *emsg, size_t emsgsize)
{
  struct tk3_data d = { .dev = TK3_NONE };
  struct tk3_channel_s *c = tk3_channel(channel, TK3_NONE, emsg, emsgsize);
  if (!c) return d;

  pthread_mutex_lock(&chanex);
  if (c->fd < 0)
    snprintf(emsg, emsgsize, "%s: disconnected", c->path);
  else
    d = c->data;
  pthread_mutex_unlock(&chanex);
  return d;
}


/* --- tk3_send_data ------------------------------------------------------- */

int
tk3_send_data(int channel, const char *data, double *vals, size_t len,
              char *emsg, size_t emsgsize)
{
  static const struct {
    enum tk3_device dev;
    const char *key, *msg;
  } dmap[] = {
    { TK3_NONE, "start", "g" },
    { TK3_NONE, "stop", "x" },
    { TK3_NONE, "vels", "w%@" },
    { TK3_MKBL, "vel", "v%@" },
    { TK3_NONE, NULL, NULL }
  };

  struct tk3_channel_s *c;
  uint16_t d[len];
  double p;
  int i, n;

  for(i = 0; dmap[i].key; i++)
    if (!strcmp(dmap[i].key, data)) break;
  if (!dmap[i].key) {
    snprintf(emsg, emsgsize, "%d: bad request", channel);
    return -1;
  }

  c = tk3_channel(channel, dmap[i].dev, emsg, emsgsize);
  if (!c) return -1;

  for(n = 0; n < len; n++) {
    p = (vals[n] > 0) ? 1e6/vals[n] : 65535;
    d[n] = (p > 65535) ? 65535 : p;
  }

  if (tk3_send_msg(c, dmap[i].msg, d, len)) {
    snprintf(emsg, emsgsize, "%s: %s", c->path, strerror(errno));
    return -1;
  }
  return 0;
}


/* --- tk3_rqst_data ------------------------------------------------------- */

int
tk3_rqst_data(int channel, const char *data, double frequency,
              char *emsg, size_t emsgsize)
{
  static const struct {
    enum tk3_device dev;
    const char *key, *msg;
  } dmap[] = {
    { TK3_MKFL, "battery", "b%4" },
    { TK3_MKFL, "imu", "i%4" },
    { TK3_NONE, "motor", "m%4" },
    { TK3_NONE, NULL, NULL }
  };

  struct tk3_channel_s *c;
  uint32_t p;
  int i;

  for(i = 0; dmap[i].key; i++)
    if (!strcmp(dmap[i].key, data)) break;
  if (!dmap[i].key) {
    snprintf(emsg, emsgsize, "%d: bad request", channel);
    return -1;
  }

  c = tk3_channel(channel, dmap[i].dev, emsg, emsgsize);
  if (!c) return -1;

  p = frequency ? 1000000/frequency : 0;
  if (tk3_send_msg(c, dmap[i].msg, p)) {
    snprintf(emsg, emsgsize, "%s: %s", c->path, strerror(errno));
    return -1;
  }
  return 0;
}


/* --- tk3_update ---------------------------------------------------------- */

static void
tk3_update(struct tk3_channel_s *c, const uint8_t *msg, size_t len)
{
  int16_t v16;
  uint16_t u16;

  switch(*msg++) {
    case 'I': /* IMU data */
      if (len == 13 && c->data.dev == TK3_MKFL) {
        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_x = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_y = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.acc_z = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_r = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_p = v16/1000.;

        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.gyr_y = v16/1000.;
      }
      break;

    case 'M': /* motor data */
      if (len == 8) {
        struct tk3_mkbl_state motor;
        uint8_t state = *msg++;

        motor.id = state & 0xf;
        motor.emerg = !!(state & 0x80);
        motor.servo = !!(state & 0x40);
        motor.spinning = !!(state & 0x20);
        motor.starting = !!(state & 0x10);

        u16 = ((uint16_t)(*msg++) << 8);
        u16 |= ((uint16_t)(*msg++) << 0);
        if (motor.spinning)
          motor.velocity = (u16 > 0) ? 1e6/u16 : 0;
        else
          motor.velocity = 0;

        u16 = ((uint16_t)(*msg++) << 8);
        u16 |= ((uint16_t)(*msg++) << 0);
        motor.pwm = u16 * 100./1024.;

        u16 = ((uint16_t)(*msg++) << 8);
        u16 |= ((uint16_t)(*msg++) << 0);
        motor.current = u16 / 1e3;

        switch(c->data.dev) {
          case TK3_MKBL:
            c->data.mkbl = motor;
            break;

          case TK3_MKFL:
            if (motor.id > c->data.mkfl.nmotor) {
              c->data.mkfl.motor =
                realloc(c->data.mkfl.motor,
                        motor.id * sizeof(*c->data.mkfl.motor));
              c->data.mkfl.nmotor = motor.id;
            }
            c->data.mkfl.motor[motor.id-1] = motor;
            break;

          case TK3_NONE: break;
        }
      }
      break;

    case 'B': /* nattery data */
      if (len == 3 && c->data.dev == TK3_MKFL) {
        v16 = ((int16_t)(*msg++) << 8);
        v16 |= ((uint16_t)(*msg++) << 0);
        c->data.mkfl.battery = v16/1000.;
      }
      break;
  }
}


/* --- tk3_reader ---------------------------------------------------------- */

static void *
tk3_reader(void *data)
{
  int i, s;

  pthread_mutex_lock(&chanex);
  while(nchannels) {
    s = tk3_wait_msg(&channels, nchannels, &chanex);

    for(i = 0; i < nchannels; i++) {
      s = tk3_recv_msg(channels + i, false);
      switch(s) {
        case 1:
          tk3_update(channels + i, channels[i].msg, channels[i].len);
          break;

        case 0:
          break;

        default:
          if (channels[i].fd >=0) close(channels[i].fd);
          channels[i].fd = -1;
      }
    }
  }
  pthread_mutex_unlock(&chanex);

  return NULL;
}
