%

%
% Copyright (c) 2015 LAAS/CNRS
% All rights reserved.
%
% Redistribution  and  use  in  source  and binary  forms,  with  or  without
% modification, are permitted provided that the following conditions are met:
%
%   1. Redistributions of  source  code must retain the  above copyright
%      notice and this list of conditions.
%   2. Redistributions in binary form must reproduce the above copyright
%      notice and  this list of  conditions in the  documentation and/or
%      other materials provided with the distribution.
%
% THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
% WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
% MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
% ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
% WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
% IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
%
%                                           Anthony Mallet on Fri Jun 26 2015
%
function tk3_block(block)
  setup(block);
end

function setup(block)
  tk3_mask(block.BlockHandle, 0);
  user = get_param(block.BlockHandle, 'UserData');

  % Register number of ports
  block.NumInputPorts  = user.inports;
  block.NumOutputPorts = user.outports;
  block.SetPreCompInpPortInfoToDynamic;
  block.SetPreCompOutPortInfoToDynamic;

  % default input/output port properties
  for i = 1:user.inports
    block.InputPort(i).Dimensions = 1;
    block.InputPort(i).DatatypeID = 0; % double
    block.InputPort(i).Complexity = 'Real';
    block.InputPort(i).SampleTime = [-1 0];
  end
  for i = 1:user.outports
    block.OutputPort(i).DimensionsMode = 'Fixed';
    block.OutputPort(i).Dimensions = 1;
    block.OutputPort(i).DatatypeID = 0; % double
    block.OutputPort(i).Complexity = 'Real';
    block.OutputPort(i).SamplingMode = 'Sample';
  end

  % specific input/output ports
  if isfield(user.outidx, 'imu')
    block.OutputPort(user.outidx.imu).Dimensions = 6;
    block.OutputPort(user.outidx.imu).SampleTime = ...
      [1/str2num(user.param.imufreq) 0];
  end
  if isfield(user.outidx, 'vel')
    block.OutputPort(user.outidx.vel).SampleTime = ...
      [1/str2num(user.param.motorfreq) 0];
  end
  if isfield(user.outidx, 'pwm')
    block.OutputPort(user.outidx.pwm).SampleTime = ...
      [1/str2num(user.param.motorfreq) 0];
  end
  if isfield(user.outidx, 'cur')
    block.OutputPort(user.outidx.cur).SampleTime = ...
      [1/str2num(user.param.motorfreq) 0];
  end
  if isfield(user.outidx, 'bat')
    block.OutputPort(user.outidx.bat).SampleTime = ...
      [1/str2num(user.param.batteryfreq) 0];
  end
  if ~isfield(user.outidx, 'mid')
    if isfield(user.inidx, 'cmd')
      block.InputPort(user.inidx.cmd).Dimensions = str2num(user.param.nmotor);
    end
    if isfield(user.outidx, 'vel')
      block.OutputPort(user.outidx.vel).Dimensions = str2num(user.param.nmotor);
    end
    if isfield(user.outidx, 'pwm')
      block.OutputPort(user.outidx.pwm).Dimensions = str2num(user.param.nmotor);
    end
    if isfield(user.outidx, 'cur')
      block.OutputPort(user.outidx.cur).Dimensions = str2num(user.param.nmotor);
    end
  end


  % Specify the block simStateCompliance. The allowed values are:
  %    'UnknownSimState', < The default setting; warn and assume DefaultSimState
  %    'DefaultSimState', < Same sim state as a built-in block
  %    'HasNoSimState',   < No sim state
  %    'CustomSimState',  < Has GetSimState and SetSimState methods
  %    'DisallowSimState' < Error out when saving or restoring the model sim state
  block.SimStateCompliance = 'HasNoSimState';

  %% -----------------------------------------------------------------
  %% The MATLAB S-function uses an internal registry for all
  %% block methods. You should register all relevant methods
  %% (optional and required) as illustrated below. You may choose
  %% any suitable name for the methods and implement these methods
  %% as local functions within the same file. See comments
  %% provided for each function for more information.
  %% -----------------------------------------------------------------

  block.RegBlockMethod('SetInputPortSamplingMode', @setinportsamplingmode);
  block.RegBlockMethod('SetInputPortSampleTime', @setinputportsampletime);
  block.RegBlockMethod('SetOutputPortSampleTime', @setoutputportsampletime);
  block.RegBlockMethod('Start', @start);
  block.RegBlockMethod('Outputs', @outputs);     % Required
  block.RegBlockMethod('Update', @update);
  block.RegBlockMethod('Terminate', @terminate); % Required
end

function setinportsamplingmode(block, idx, fd)
  block.InputPort(idx).SamplingMode = fd;
end

function setinputportsampletime(block, idx, time)
  block.InputPort(idx).SampleTime = time;
end
function setoutputportsampletime(block, idx, time)
  % outport frequency should not be changed
end

%%
%% Start:
%%   Functionality    : Called once at start of model execution. If you
%%                      have states that should be initialized once, this
%%                      is the place to do it.
%%   Required         : No
%%   C-MEX counterpart: mdlStart
%%
function start(block)
  user = get_param(block.BlockHandle, 'UserData');
  switch user.param.device
    case 'Flight controller'
      dev = 'mkfl';
    case 'Brushless controller'
      dev = 'mkbl';
  end
  user.client = tk3.client(dev, ...
                           user.param.serial, str2num(user.param.speed));

  if isfield(user.outidx, 'imu')
    user.client.get_imu(str2num(user.param.imufreq));
  end
  if isfield(user.outidx, 'vel')
    user.client.get_motor(str2num(user.param.motorfreq));
  end
  if isfield(user.outidx, 'bat')
    user.client.get_battery(str2num(user.param.batteryfreq));
  end
  user.client.stop;
  user.start = 0;

  set_param(block.BlockHandle, 'UserData', user);
end

function outputs(block)
  user = get_param(block.BlockHandle, 'UserData');
  if length(fieldnames(user.outidx)) > 0
    data = user.client.data;
  end

  if isfield(user.outidx, 'imu')
    block.OutputPort(user.outidx.imu).Data = ...
    [ data.acc_x, data.acc_y, data.acc_z, ...
      data.gyr_r, data.gyr_p, data.gyr_y ];
  end

  if isfield(user.outidx, 'mid')
    block.OutputPort(user.outidx.mid).Data = data.id;
    if isfield(user.outidx, 'vel')
      block.OutputPort(user.outidx.vel).Data = data.velocity;
    end
    if isfield(user.outidx, 'pwm')
      block.OutputPort(user.outidx.pwm).Data = data.pwm;
    end
    if isfield(user.outidx, 'cur')
      block.OutputPort(user.outidx.cur).Data = data.current;
    end
  else
    if isfield(user.outidx, 'vel')
      for i = 1:length(block.OutputPort(user.outidx.vel).Data)
        if i > length(data.motor)
           block.OutputPort(user.outidx.vel).Data(i) = NaN;
        else
          block.OutputPort(user.outidx.vel).Data(i) = data.motor{i}.velocity;
        end
      end
    end
    if isfield(user.outidx, 'pwm')
      for i = 1:length(block.OutputPort(user.outidx.pwm).Data)
        if i > length(data.motor)
          block.OutputPort(user.outidx.pwm).Data(i) = NaN;
        else
          block.OutputPort(user.outidx.pwm).Data(i) = data.motor{i}.pwm;
        end
      end
    end
    if isfield(user.outidx, 'cur')
      for i = 1:length(block.OutputPort(user.outidx.cur).Data)
        if i > length(data.motor)
          block.OutputPort(user.outidx.cur).Data(i) = NaN;
        else
          block.OutputPort(user.outidx.cur).Data(i) = data.motor{i}.current;
        end
      end
    end
  end

  if isfield(user.outidx, 'bat')
    block.OutputPort(user.outidx.bat).Data = data.battery;
  end
end

function update(block)
  user = get_param(block.BlockHandle, 'UserData');

  if isfield(user.inidx, 'start')
    n = block.InputPort(user.inidx.start).Data;
    if user.start ~= n
       user.start = n;
       if user.start
          user.client.start;
       else
          user.client.stop;
       end
       set_param(block.BlockHandle, 'UserData', user);
    end
  end
  if isfield(user.inidx, 'cmd')
    user.client.velocity(block.InputPort(user.inidx.cmd).Data);
  end
end

function terminate(block)
  user = get_param(block.BlockHandle, 'UserData');
  if isfield(user, 'client')
     delete(user.client);
  end
end
